base_mappings = {
    "curve_mappings": {
        "4A_PRES": "4A_PRESS",
        "A16L": "RSHA",
        "A28H": "RMED",
        "A40L": "RDEP",
        "AC": "AC",
        "ACS": "ACS",
        "AI": "AI",
        "AI10": "RSHA",
        "AI60": "RMED",
        "AI90": "RDEP",
        "AO10": "RSHA",
        "AO60": "RMED",
        "AO90": "RDEP",
        "APLC": "APLC",
        "AT10": "RSHA",
        "AT60": "RMED",
        "AT90": "RDEP",
        "BFV": "BFV",
        "BFV_UHR": "BFV",
        "BPHI": "BPHI",
        "BS": "BS",
        "CAL": "CALI",
        "CALI": "CALI",
        "CALS": "CALI",
        "CGR": "CGR",
        "CMRT": "MRP",
        "CNC": "NEU",
        "DCAV": "CALI",
        "DCOR": "DENC",
        "DEN": "DEN",
        "DENC": "DENC",
        "DEN_VIRGIN": "DEN_VIRGIN",
        "DRHO": "DENC",
        "DT": "AC",
        "DTC": "AC",
        "DTCO": "AC",
        "DTS": "ACS",
        "DTSM": "ACS",
        "FFV": "FFV",
        "FFV_UHR": "FFV",
        "FPRESS": "FPRESS",
        "GR": "GR",
        "GRMA": "GR",
        "GR_EDTC": "GR",
        "HAC": "AC",
        "HCAL": "CALI",
        "HCGR": "CGR",
        "HCNC": "NEU",
        "HCNL": "NEU",
        "HCOR": "DENC",
        "HCORR": "DENC",
        "HDAR": "CALI",
        "HDEN": "DEN",
        "HDRA": "DENC",
        "HDRH": "DENC",
        "HDT": "AC",
        "HFK": "K",
        "HGR": "GR",
        "HPHI": "NEU",
        "HRD": "RDEP",
        "HRHO": "DEN",
        "HRM": "RMED",
        "HRS": "RMIC",
        "HSGR": "SGR",
        "HTHO": "TH",
        "HURA": "U",
        "IDDR": "DENC",
        "IDPE": "PEF",
        "IDRO": "DEN",
        "ILD": "RDEP",
        "ILM": "RMED",
        "K": "K",
        "LLD": "RDEP",
        "LLS": "RSHA",
        "M": "M",
        "MRP": "MRP",
        "MSFL": "RMIC",
        "NEU": "NEU",
        "NPHI": "NEU",
        "NPOR": "NEU",
        "P16L": "RSHA",
        "P28L": "RMED",
        "P40H": "RDEP",
        "P40L": "RDEP",
        "PE": "PEF",
        "PEB": "PEF",
        "PEF": "PEF",
        "PEFZ": "PEF",
        "POTA": "K",
        "RD": "RDEP",
        "RDEP": "RDEP",
        "RH72_1DF": "RH",
        "RHGA": "RHGE",
        "RHGE": "RHGE",
        "RHO8": "DEN",
        "RHOB": "DEN",
        "RHOZ": "DEN",
        "RMED": "RMED",
        "RMIC": "RMIC",
        "RMLL": "RMIC",
        "ROBB": "DEN",
        "ROP": "ROP",
        "RS": "RSHA",
        "RSHA": "RSHA",
        "RT": "RT",
        "RT_HRLT": "RDEP",
        "RV72_1DF": "RV",
        "RXO": "RXO",
        "RXOZ": "RXO",
        "RXO_ADT": "RXO",
        "RXO_HRLT": "RXO",
        "SFLU": "SFLU",
        "SGR": "SGR",
        "SHAR": "RSHA",
        "SP": "SP",
        "SPBR": "SP",
        "TEMP": "TEMP",
        "TH": "TH",
        "THOR": "TH",
        "TNPH": "NEU",
        "U": "U",
        "UCAV": "CALI",
        "URAN": "U",
        "WANH": "WANH",
        "WCAR": "WCAR",
        "WCLA": "WCLA",
        "WCLC": "WCLC",
        "WCOA": "WCOA",
        "WDOL": "WDOL",
        "WEVA": "WEVA",
        "WPYR": "WPYR",
        "WQFM": "WQFM",
        "WSID": "WSID",
        "ZCOR": "DENC",
        "ZDEN": "DEN",
        "ZONE": "lsuName",
        "lsuName": "lsuName",
        "wellName": "well_name",
        "well_name": "well_name",
        "WELL": "well_name",
        "well": "well_name",
    }
}
